/* ------------------------------------------------------------------------
 *	Nixterm  -  basic terminal for LibGGI.
 * ------------------------------------------------------------------------
 *
 *	Copyright (c) 1998-1999  Andrew Apted  <ajapted@netspace.net.au>
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License as
 *	published by the Free Software Foundation; either version 2, or
 *	(at your option) any later version.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; see the file `COPYING'.   If not, write
 *	to the Free Software Foundation, Inc., 59 Temple Place - Suite
 *	330, Boston, MA 02111-1307, USA
 *
 * ------------------------------------------------------------------------
 *
 *	NOTES
 *
 *	-  This is a small terminal program for LibGGI.  As the name
 *	   suggests, it only offers a minimal amount of functionality.
 *	   One feature is the scrollback support (shift + PGUP / PGDN
 *	   as in Linux).  Another feature is the blanking after a period
 *	   of inactivity.
 *
 *	-  Nixterm has it's own set of control codes (hey, where's the
 *	   fun in being XXX or YYY compatible ? :-).  Thus you need to
 *	   run "tic nixterm.terminfo" as root to install the terminfo
 *	   entry (NOTE: this is done by `make install').  Programs that
 *	   use ncurses (e.g. lynx, vim, mutt, midnight commander) should
 *	   run satisfactorily.
 *
 * ------------------------------------------------------------------------
 *
 *	TODO
 *
 *	-  implement visual bell.
 *	-  speed up text using ggiPuts() for multible chars.
 *	-  use ggiFlushRegion() for better X performance.
 *
 * ------------------------------------------------------------------------
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <signal.h>
#include <termios.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <fcntl.h>

#include <ggi/ggi.h>


#define VERSION_STR	"version 2.0 (c) 1998-1999 Andrew Apted."


/* Macros */

#ifndef MIN
#define MIN(a,b)  ((a) < (b) ? (a) : (b))
#endif
#ifndef MAX
#define MAX(a,b)  ((a) > (b) ? (a) : (b))
#endif
#ifndef ABS
#define ABS(x)  (((x) < 0) ? -(x) : (x))
#endif


/* Global Vars */

static ggi_visual_t vis;
static ggi_mode vis_mode;

static int shell_pid;

static volatile int quit=0;

#define SLEEP_MS  20

#define WANTED_EVENTS  (emKey | emExpose)


#define WRITE_MAX  4096

static unsigned char write_buf[WRITE_MAX];
static int write_pending = 0;


typedef unsigned short  nix_char;

#define NIX_MASK_CHAR    0x00ff
#define NIX_MASK_ATTR    0xff00

#define NIX_ATTR_BOLD    0x0100
#define NIX_ATTR_REV     0x0200
#define NIX_ATTR_S_OUT   0x0400
#define NIX_ATTR_ULINE   0x0800
#define NIX_ATTR_DIM     0x1000
#define NIX_ATTR_INVIS   0x2000


static nix_char *buffer;

static ggi_coord screen;	/* window size */
static ggi_coord cursor;	/* cursor position */
static ggi_coord font;		/* character size */

static int win_top;		/* window position */
static int scroll_top;		/* scroll region top */
static int scroll_bottom;	/* scroll region bottom */

static int scrollback = 200;
static int sb_step = 5;

static int blank_time = 5 * 60 * 1000;	/* milliseconds */
static int blanked;
static struct timeval blank_start;


#define NIX_OPT_QUIET      0x0001
#define NIX_OPT_REVERSE    0x0002
#define NIX_OPT_COLOR      0x0004


static int cur_attr=0;
static int options=0;
static int cursor_enable=1;
static ggi_coord saved_cursor;

static int parse_state=0;    /* for partial ^Y, can be 1..2 */
static int parse_1, parse_2;


static ggi_color palette[16] =
{
	{ 0x0000, 0x0000, 0x0000 },
	{ 0x0000, 0x0000, 0xaaaa },
	{ 0x0000, 0xaaaa, 0x0000 },
	{ 0x0000, 0xaaaa, 0xaaaa },
	{ 0xaaaa, 0x0000, 0x0000 },
	{ 0xaaaa, 0x0000, 0xaaaa },
	{ 0xaaaa, 0xaaaa, 0x0000 },
	{ 0xaaaa, 0xaaaa, 0xaaaa },

	{ 0x5555, 0x5555, 0x5555 },
	{ 0x5555, 0x5555, 0xffff },
	{ 0x5555, 0xffff, 0x5555 },
	{ 0x5555, 0xffff, 0xffff },
	{ 0xffff, 0x5555, 0x5555 },
	{ 0xffff, 0x5555, 0xffff },
	{ 0xffff, 0xffff, 0x5555 },
	{ 0xffff, 0xffff, 0xffff }
};

static ggi_pixel lookup[16];


/* ---------------------------------------------------------------------- */


static void map_colors(void)
{
	int i;

	for (i=0; i < 16; i++) {
		lookup[i] = ggiMapColor(vis, palette + i);
	}
}

static void blacken_whole_screen(void)
{
	ggiSetGCForeground(vis, lookup[0]);
	ggiFillscreen(vis);
}

static int calc_colors(int attr)
{
	/* Returns: (background << 8) | foreground.
	 */

	if (! (options & NIX_OPT_COLOR)) {

		if (attr & NIX_ATTR_INVIS) {
			return 0x0000;
		
		} else if (attr & (NIX_ATTR_REV | NIX_ATTR_S_OUT)) {
			return 0x0700;

		} else if (attr & (NIX_ATTR_BOLD | NIX_ATTR_ULINE)) {
			return 0x000f;
		}

		return 0x0007;
	}
	
	/* color */

	if (attr & NIX_ATTR_INVIS) {
		return 0x0000;
		
	} else if (attr & NIX_ATTR_REV) {
		return 0x010b;   /* bright cyan on dark blue */

	} else if (attr & NIX_ATTR_S_OUT) {
		return 0x040e;   /* bright yellow on dark red */

	} else if (attr & NIX_ATTR_BOLD) {
		return 0x000e;   /* bright yellow */
		
	} else if (attr & NIX_ATTR_ULINE) {
		return 0x000a;   /* bright green */
		
	} else if (attr & NIX_ATTR_DIM) {
		return 0x000d;   /* bright purple */
	}

	return 0x0007;
}

static void draw_char(int sx, int sy, nix_char c, int is_cursor)
{
	int bgfg = calc_colors(c & NIX_MASK_ATTR);

	ggi_pixel bg = lookup[(bgfg >> 8) & 0x0f];
	ggi_pixel fg = lookup[bgfg & 0x0f];

	if (is_cursor) {
		if ((options & NIX_OPT_COLOR) && 
		    (bg != lookup[4]) && (fg != lookup[4])) {
			bg = lookup[4];
		} else {
			ggi_pixel tmp=bg; bg=fg; fg=tmp;
		}
	}

	ggiSetGCBackground(vis, bg);
	ggiSetGCForeground(vis, fg);

	ggiPutc(vis, sx*font.x, sy*font.y, c & NIX_MASK_CHAR);
}

static void do_putc(int sx, int sy, nix_char c)
{
	int actual_y = cursor.y + scrollback - win_top;
	
	/* update visible area */

	if (actual_y < screen.y) {

		draw_char(sx, actual_y, c, 0);
	}

	/* update buffer */

	buffer[(scrollback + sy)*screen.x + sx] = c;
}

static void do_redraw(int sx, int sy, int sw, int sh)
{
	int x, y;

	if ((sw <= 0) || (sh <= 0)) return;

	for (y=sy; y < sy+sh; y++)
	for (x=sx; x < sx+sw; x++) {

		draw_char(x, y, buffer[(win_top+y)*screen.x + x], 0);
	}
}

static void do_clear(int x, int y, int w, int h)
{
	int sy = y - win_top;
	int sh = h;
	
	if ((w <= 0) || (h <= 0)) return;


	/* update buffer */
	
	for (; h > 0; h--, y++) {
		memset(buffer + y*screen.x + x, 0, w*sizeof(nix_char));
	}


	/* update visible area */

	if (sy < 0) {
		sh += sy; sy = 0;
	}
	if (sy+sh > screen.y) {
		sh = screen.y - sy;
	}

	if (sh > 0) {
		ggiSetGCForeground(vis, lookup[0]);
		ggiDrawBox(vis, x*font.x, sy*font.y, w*font.x, sh*font.y);
	}
}

static void do_move(int x, int y, int nx, int ny, int w, int h)
{
	int sy = y  - win_top;
	int ty = ny - win_top;
	int dy;

	int tot_h = h;
	int tfr_h;
	
	if ((w <= 0) || (h <= 0)) return;


	/* update buffer */

	if (ny > y) {
		for (; h > 0; h--) {
			memmove(buffer + (ny+h-1)*screen.x + nx,
				buffer + ( y+h-1)*screen.x + x,
				w * sizeof(nix_char));
		}
	} else {
		for (; h > 0; h--) {
			memmove(buffer + (ny)*screen.x + nx,
				buffer + ( y)*screen.x + x,
				w * sizeof(nix_char));
			y++; ny++;
		}
	}


	/* update visible area */

	if (ty < 0) {
		tot_h += ty; ty = 0;
	}
	if (ty+tot_h > screen.y) {
		tot_h = screen.y - ty;
	}

	if (tot_h <= 0) return;
		

	dy    = ty; 
	tfr_h = tot_h;

	if (sy < 0) {
		tfr_h += sy; dy -= sy; sy = 0;
	}
	if (sy+tfr_h > screen.y) {
		tfr_h = screen.y - sy;
	}

	
	/* do transfer */

	if (tfr_h > 0) {
		ggiCopyBox(vis, x*font.x,  sy*font.y, 
				w*font.x,  tfr_h*font.y,
				nx*font.x, dy*font.y);
	}

	/* redraw any parts not transferred */

	if (tfr_h <= 0) {
		do_redraw(nx, ty, w, tot_h);
		return;
	}
	
	if (ty < dy) {
		do_redraw(nx, ty, w, MIN(dy-ty, tot_h));
	}

	if (ty+tot_h > dy+tfr_h) {
		int new_h = MIN(ty+tot_h-(dy+tfr_h), tot_h);
		do_redraw(nx, ty+tot_h-new_h, w, new_h);
	}
}

static void add_line_to_scrollback(void)
{
	if (scrollback == 0) {
		return;
	}

	/* scroll up scrollback area */

	do_move(0, 1, 0, 0, screen.x, scrollback-1);

	/* copy top line */

	do_move(0, scrollback, 0, scrollback-1, screen.x, 1);
}

static void do_cursor(int state)
{
	int actual_y = cursor.y + scrollback - win_top;

	if (actual_y >= screen.y)
		return;

	draw_char(cursor.x, actual_y,
		  buffer[(scrollback + cursor.y)*screen.x + cursor.x],
		  cursor_enable ? state : 0);
}

static void do_expose(int x, int y, int w, int h)
{
	/* redraw terminal */

	do_cursor(0);
	do_redraw(x, y, w, h);
	do_cursor(1);
	
	ggiFlush(vis);
}


/* ---------------------------------------------------------------------- */


static void visual_beep(void)
{
	/* !!! */
}

static void beep(void)
{
	int fd;
	unsigned char buf[4];
	
	if (options & NIX_OPT_QUIET) {
		visual_beep();
		return;
	}

	/* try opening /dev/tty and sending ^G */

	if ((fd = open("/dev/tty", O_WRONLY)) < 0) {
		/* failed */
		visual_beep();
		return;
	}

	buf[0] = 7;  /* ^G */
	
	write(fd, buf, 1);
	close(fd);
}

static void do_wrapping(void)
{
	if (cursor.x < 0) {
		cursor.x = 0;
	}

	if (cursor.x >= screen.x) {
		cursor.x = 0; 
		cursor.y += 1;
	}

	if (scroll_top >= scroll_bottom) {
		return;
	}

	while (cursor.y < scroll_top) {

		/* scroll down */

		do_move(0, scrollback + scroll_top, 
			0, scrollback + scroll_top + 1, 
			screen.x, scroll_bottom - scroll_top - 1);

		do_clear(0, scrollback + scroll_top, screen.x, 1);

		cursor.y++;
	}

	while (cursor.y >= scroll_bottom) {

		/* scroll up */

		if (scroll_top == 0) {
			add_line_to_scrollback();
		}
		
		do_move(0, scrollback + scroll_top + 1, 
			0, scrollback + scroll_top,
			screen.x, scroll_bottom - scroll_top - 1);

		do_clear(0, scrollback + scroll_bottom - 1, screen.x, 1);

		cursor.y--;
	}
}

static void handle_Y_escape(void)
{
	switch (parse_1) {

		case 'H':   /* hpa */
		{
			int arg = (parse_2 - ' ') & 0xff;

			cursor.x = MIN(arg, screen.x-1);
			return;
		}

		case 'V':   /* vpa */
		{
			int arg = (parse_2 - ' ') & 0xff;

			cursor.y = MIN(arg, screen.y-1);
			return;
		}

		case 'T':   /* csr (top) */
		{
			int arg = (parse_2 - ' ') & 0xff;

			scroll_top = MIN(arg, screen.y-1);
			return;
		}
		
		case 'B':   /* csr (bottom) */
		{
			int arg = (parse_2 - ' ') & 0xff;

			scroll_bottom = 1 + MIN(arg, screen.y-1);
			return;
		}
		
		case 'C':   /* civis, cnorm, sc & rc */
		{
			switch (parse_2) {
			
				case '0': cursor_enable = 0; break;
				case '1': cursor_enable = 1; break;

				case '>': saved_cursor = cursor; break;
				case '<': cursor = saved_cursor; break;
			}
			return;
		}
		
		case 'R':   /* reset */
		{
			if ((parse_2 == '1') || (parse_2 == '2')) {
				
				/* soft reset */

				cur_attr = 0;
				cursor_enable = 1;
				
				scroll_top = 0;
				scroll_bottom = screen.y;
			}

			if (parse_2 == '2') {
				
				/* hard reset */

				do_clear(0, 0, screen.x, 
					 scrollback + screen.y);

				cursor.x = 0;
				cursor.y = 0;
			}

			return;
		}
		
		case '+':   /* set attribute */
		case '-':   /* reset attribute */
		{
			int flag=0;

			switch (parse_2) {

				case 'B': flag=NIX_ATTR_BOLD;  break;
				case 'R': flag=NIX_ATTR_REV;   break;
				case 'S': flag=NIX_ATTR_S_OUT; break;
				case 'U': flag=NIX_ATTR_ULINE; break;
				case 'D': flag=NIX_ATTR_DIM;   break;
				case 'I': flag=NIX_ATTR_INVIS; break;

				/* set/reset All attributes */
				case 'A': flag=NIX_MASK_ATTR; break;
			}

			if (parse_1 == '+') {
				cur_attr |= flag;
			} else {
				cur_attr &= ~flag;
			}
			return;
		}
	}
}

static void print_char(int c)
{
	if (parse_state == 1) {
		parse_1 = c;
		parse_state++;
		return;
	}

	if (parse_state == 2) {
		parse_2 = c;
		handle_Y_escape();
		parse_state = 0;
		return;
	}
	
	switch(c) {
		case 0:	  /* NUL: do nothing */
		{
			return;
		}
			
		case 7:   /* ^G: bel */
		{
			beep();
			return;
		}

		case 8:   /* ^H: cub1 */
		{
			cursor.x--;
			do_wrapping(); 
			return;
		}

		case 14:   /* ^N: cuf1 */
		{
			cursor.x++;
			do_wrapping(); 
			return;
		}

		case 9:   /* ^I: ht */
		{
			cursor.x += (8 - (cursor.x & 0x07));
			do_wrapping(); 
			return;
		}

		case 1:   /* ^A: ind */
		case 10:  /* ^J: cud1 */
		{
			cursor.y++;
			do_wrapping(); 
			return;
		}

		case 2:   /* ^B: ri */
		case 11:  /* ^K: cuu1 */
		{
			cursor.y--;
			do_wrapping(); 
			return;
		}

		case 13:  /* ^M: cr */
		{
			cursor.x = 0; 
			return;
		}

		case 12:  /* ^L: clear */
		{
			do_clear(0, scrollback, screen.x, screen.y);

			/* fall through... */
		}
		case 5:   /* ^E: home */
		{
			cursor.x = 0;
			cursor.y = 0;
			return;
		}

		case 6:   /* ^F: ll */
		{
			cursor.x = 0;
			cursor.y = screen.y - 1;
			return;
		}

		case 20:   /* ^T: il1 */
		{
			if ((cursor.y < scroll_top) ||
			    (cursor.y >= scroll_bottom)) return;

			do_move(0, scrollback + cursor.y, 
				0, scrollback + cursor.y + 1,
				screen.x, scroll_bottom - cursor.y - 1);

			do_clear(0, scrollback + cursor.y, screen.x, 1);
			return;
		}

		case 21:   /* ^U: dl1 */
		{
			if ((cursor.y < scroll_top) ||
			    (cursor.y >= scroll_bottom)) return;

			do_move(0, scrollback + cursor.y + 1, 
				0, scrollback + cursor.y,
				screen.x, scroll_bottom - cursor.y - 1);

			do_clear(0, scrollback + scroll_bottom - 1, 
				 screen.x, 1);
			return;
		}

		case 23:   /* ^W: ich1 */
		{
			do_move(cursor.x,      scrollback + cursor.y, 
				 cursor.x + 1, scrollback + cursor.y,
				 screen.x - cursor.x - 1, 1);

			do_clear(cursor.x, scrollback + cursor.y, 1, 1);
			return;
		}

		case 24:   /* ^X: dch1 */
		{
			do_move(cursor.x + 1, scrollback + cursor.y, 
				 cursor.x,    scrollback + cursor.y,
				 screen.x - cursor.x - 1, 1);

			do_clear(screen.x - 1, scrollback + cursor.y, 1, 1);
			return;
		}

		case 15:   /* ^O: el */
		{
			do_clear(cursor.x, scrollback + cursor.y, 
				 screen.x - cursor.x, 1);
			return;
		}
		
		case 16:   /* ^P: ed */
		{
			do_clear(0, scrollback + cursor.y, 
				 screen.x, screen.y - cursor.y);
			return;
		}
		
		case 25:   /* ^Y: nixterm escape */
		{
			parse_state=1;
			return;
		}
	}

	if ((c < 0x20) || (c > 0x7e)) {
		c = '?';
	}

	do_putc(cursor.x, cursor.y, cur_attr | c);

	cursor.x++;
	do_wrapping();
}

static void move_window(int delta_y)
{
	win_top += delta_y;

	if (win_top < 0)	  win_top = 0;
	if (win_top > scrollback) win_top = scrollback;

	do_expose(0, 0, screen.x, screen.y);   /* cheap 'n' nasty */
}

static void do_blank(int state)
{
	if (! blanked && state) {		
	
		/* blanking */

		blacken_whole_screen();

	} else if (blanked && ! state) {	
	
		/* unblanking */

		do_expose(0, 0, screen.x, screen.y);
	}

	blanked = state;
	gettimeofday(&blank_start, NULL);
}

static void send_char(int fd, int c)
{
	if (write_pending >= WRITE_MAX) {

		/* output pipe is full */
		beep();
		return;
	}

	write_buf[write_pending++] = c;
}

static void send_str(int fd, unsigned char *s)
{
	/* Another cheap shot... */

	while (*s) {
		send_char(fd, *s++);
	}
}

static void handle_keyboard(int sym, int modifiers, int out_fd)
{
	int shift = modifiers & (1 << GII_KM_SHIFT);
	
	if (GII_KTYP(sym) == GII_KT_LATIN1) {
		send_char(out_fd, GII_KVAL(sym)); 
		return;
	}

	if (shift) switch (sym) {

		case GIIK_PageUp:   sym = GIIK_ScrollBack; break;
		case GIIK_PageDown: sym = GIIK_ScrollForw; break;

		case GIIK_F1: sym = GIIK_F11; break;
		case GIIK_F2: sym = GIIK_F12; break;
	}

	switch (sym) {

		/* keys we pass on */

		case GIIK_Enter: send_char(out_fd, '\n'); return;
		case GIIK_Up:    send_str(out_fd, "\033[A"); return;
		case GIIK_Down:  send_str(out_fd, "\033[B"); return;
		case GIIK_Right: send_str(out_fd, "\033[C"); return;
		case GIIK_Left:  send_str(out_fd, "\033[D"); return;

		case GIIK_Home:     send_str(out_fd, "\033[1~"); return;
		case GIIK_End:      send_str(out_fd, "\033[4~"); return;
		case GIIK_Insert:   send_str(out_fd, "\033[2~"); return;
/* ???		case GIIK_Remove:   send_str(out_fd, "\033[3~"); return; */
		case GIIK_PageUp:   send_str(out_fd, "\033[5~"); return;
		case GIIK_PageDown: send_str(out_fd, "\033[6~"); return;

		case GIIK_F1:  send_str(out_fd, "\033[[A"); return;
		case GIIK_F2:  send_str(out_fd, "\033[[B"); return;
		case GIIK_F3:  send_str(out_fd, "\033[[C"); return;
		case GIIK_F4:  send_str(out_fd, "\033[[D"); return;
		case GIIK_F5:  send_str(out_fd, "\033[[E"); return;
		case GIIK_F6:  send_str(out_fd, "\033[17~"); return;
		case GIIK_F7:  send_str(out_fd, "\033[18~"); return;
		case GIIK_F8:  send_str(out_fd, "\033[19~"); return;
		case GIIK_F9:  send_str(out_fd, "\033[20~"); return;
		case GIIK_F10: send_str(out_fd, "\033[21~"); return;

		/* keys we greedily keep to ourselves */

		case GIIK_F11: quit=1; return;

		case GIIK_ScrollForw: move_window( sb_step); return;
		case GIIK_ScrollBack: move_window(-sb_step); return;
	}
}


/* ---------------------------------------------------------------------- */


static void handle_incoming_data(int fd)
{
	unsigned char buf[1024];

	int num, i;

	do {
		num = read(fd, buf, 1024);

	} while ((num < 0) && (errno = EINTR) && !quit);

	if (num == 0) {
		fprintf(stderr, "nixterm: read: EOF\n");
		quit=3; return;

	} else if (num < 0) {
		perror("nixterm: read");
		quit=3; return;
	}
	
	for (i=0; i < num; i++) {
		do_cursor(0);
		print_char(buf[i]);
		do_cursor(1);
	}

	ggiFlush(vis);
}

static void handle_outgoing_data(int fd)
{
	int num;

	do {
		num = write(fd, write_buf, write_pending);

	} while ((num < 0) && (errno = EINTR) && !quit);

	if (num < 0) {
		perror("nixterm: write");
		quit=3; return;
	}

	if ((0 < num) && (num < write_pending)) {
		memmove(write_buf, write_buf + num, write_pending - num);
	}

	write_pending -= num;
}

static void child_died(dummy)
{
	quit=2;
}

static int find_free_tty(int *master_fd, int *slave_fd)
{
	int n;

	char master_name[16];
	char slave_name[16];

	for (n=0; n < 16; n++) {

		sprintf(master_name, "/dev/ptyp%x", n);

		*master_fd = open(master_name, O_RDWR);

		if (*master_fd >= 0) {
			/* found it */
			break;
		}
	}

	if (n == 16) {
		return -1;   /* not found */
	}

	
	sprintf(slave_name, "/dev/ttyp%x", n);

	*slave_fd = open(slave_name, O_RDWR);

	if (*slave_fd < 0) {
		perror(slave_name);
		close(*master_fd);
		return -1;
	}

	return 0;
}

static void usage(void)
{
	fprintf(stderr,
		"\n"
		"USAGE: nixterm [OPTIONS...]\n"
		"\n"
		"where option is one of:\n"
		"\n"
		"      -h  --help\n"
		"      -V  --version\n"
		"      -m  --mode        <mode-string>\n"
		"      -t  --target      <target-string>\n"
		"      -i  --input       <input-string>\n"
		"      -f  --font        <width>x<height>\n"
		"      -s  --scrollback  <num-lines>\n"
		"      -b  --blank       <minutes>\n"
		"      -c  --color\n"
		"      -q  --quiet\n"
		"      -r  --reverse\n"
		"\n");
}

static int parse_font(char *s, ggi_coord *f)
{
	int w, h;

	if ((sscanf(s, " %d x %d", &w, &h) != 2) ||
	    (w <= 0) || (h <= 0)) {
		
		fprintf(stderr, "nixterm: Bad font size '%s'.\n", s);
		return -1;
	}

	f->x = w; 
	f->y = h;

	return 0;
}


/* ---------------------------------------------------------------------- */


#define CMP_OPT(short,long,extra)  \
    ((argc > extra) && \
     ((strcmp(*argv, short) == 0) || (strcmp(*argv, long) == 0)))

int main(int argc, char **argv, char **envp)
{
	char *target = NULL;
	char *inputs = NULL;
	char *shell  = "/bin/sh";

	struct winsize window_size;

	int master_fd, slave_fd;
	int w, h;
	int err;


	/* initialize */

	ggiParseMode("", &vis_mode);
	
	if (getenv("SHELL") != NULL) {
		shell = getenv("SHELL");
	}


	/* handle arguments */

	argv++; argc--;

	while (argc > 0) {

		if (CMP_OPT("-h", "--help", 0)) {

			usage();
			return 0;
		}

		if (CMP_OPT("-V", "--version", 0)) {

			printf("Nixterm %s\n", VERSION_STR);
			return 0;
		}

		if (CMP_OPT("-m", "--mode", 1)) {
		    
			ggiParseMode(argv[1], &vis_mode);
			argv += 2; argc -= 2;
			continue;
		}

		if (CMP_OPT("-t", "--target", 1)) {

			target = argv[1];
			argv += 2; argc -= 2;
			continue;
		}

		if (CMP_OPT("-i", "--input", 1)) {

			inputs = argv[1];
			argv += 2; argc -= 2;
			continue;
		}

		if (CMP_OPT("-f", "--font", 0)) {

			if (parse_font(argv[1], &font) < 0) {
				return 1;
			}
			argv += 2; argc -= 2;
			continue;
		}

		if (CMP_OPT("-s", "--scrollback", 0)) {

			if ((sscanf(argv[1], "%d", &scrollback) != 1) ||
			    (scrollback < 0)) {

				fprintf(stderr, "nixterm: Bad scrollback.\n");
				return 1;
			}
			argv += 2; argc -= 2;
			continue;
		}

		if (CMP_OPT("-b", "--blank_time", 0)) {

			if ((sscanf(argv[1], "%d", &blank_time) != 1) ||
			    (blank_time < 0)) {

				fprintf(stderr, "nixterm: Bad blank_time.\n");
				return 1;
			}
			blank_time = blank_time * 60 * 1000;
			argv += 2; argc -= 2;
			continue;
		}

		if (CMP_OPT("-q", "--quiet", 0)) {

			options |= NIX_OPT_QUIET; 
			argv++; argc--;
			continue;
		}

		if (CMP_OPT("-r", "--reverse", 0)) {

			options |= NIX_OPT_REVERSE;
			argv++; argc--;
			continue;
		}

		if (CMP_OPT("-c", "--color", 0)) {

			options |= NIX_OPT_COLOR; 
			argv++; argc--;
			continue;
		}

		fprintf(stderr, "nixterm: Unknown option '%s'\n", *argv);
		return 1;
	}


	/* now setup screen */

	if (ggiInit() < 0) {
		fprintf(stderr, "nixterm: ggiInit() failed.\n");
		return 1;
	}

	vis = ggiOpen(target, NULL);

	if (vis == NULL) {
		fprintf(stderr, "nixterm: ggiOpen() failed.\n");
		ggiExit();
		return 1;
	}

	ggiCheckMode(vis, &vis_mode);

	err = ggiSetMode(vis, &vis_mode);

	if (err) {
		fprintf(stderr,"nixterm: Unable to set mode.\n");
		return 2;
	}

	ggiSetFlags(vis, GGIFLAG_ASYNC);

	if (GT_SCHEME(vis_mode.graphtype) == GT_PALETTE) {
		ggiSetColorfulPalette(vis);
	}

/*	ggiSetCharSize(vis, font.x, font.y);	*/
	ggiGetCharSize(vis, &w, &h);

	font.x = w; font.y = h;

	screen.x = vis_mode.virt.x / font.x;
	screen.y = vis_mode.virt.y / font.y;

	buffer = calloc(screen.x * (scrollback+screen.y), sizeof(nix_char));

	cursor.x = 0;
	cursor.y = 0;

	saved_cursor = cursor;
	win_top = scrollback;

	scroll_top = 0;
	scroll_bottom = screen.y;

	blanked = 0;
	gettimeofday(&blank_start, NULL);


	/* add extra inputs */

	if (inputs) {
		ggiJoinInputs(vis, giiOpen(inputs, NULL));
	}
	

	/* map colors */

	map_colors();

	ggiSetGCForeground(vis, lookup[0]);
	ggiFillscreen(vis);


	/* create pipes */

	if (find_free_tty(&master_fd, &slave_fd) < 0) {
		fprintf(stderr, "nixterm: couldn't open PTY.\n");
		return 1;
	}
	
	setenv("TERM", "nixterm", 1);

	signal(SIGCHLD, child_died);
	signal(SIGWINCH, SIG_IGN);
	

	/* set size of terminal */

	window_size.ws_col = screen.x;
	window_size.ws_row = screen.y;
	window_size.ws_xpixel = screen.x * font.x;
	window_size.ws_ypixel = screen.y * font.y;
	
	if (ioctl(master_fd, TIOCSWINSZ, &window_size) < 0) {
		perror("nixterm: Error setting window size");
	}
	
	
	/* now spawn a shell */

	if ((shell_pid = fork()) == 0) {

		char *argv[] = { shell, NULL };

		close(0);
		close(1);
		close(2);

		close(master_fd);

		dup2(slave_fd, 0);
		dup2(slave_fd, 1);
		dup2(slave_fd, 2);

		/* set user/group/session ids */

		seteuid(getuid());
		setegid(getgid());

		setsid();

		if (ioctl(slave_fd, TIOCSCTTY, 0) < 0) {
			perror("nixterm: ioctl TIOCSCTTY");
		}

		execve(argv[0], argv, envp);

		exit(1);
	}

	close(slave_fd);


	/* main loop */

	for (;;) {
		fd_set rfds;
		fd_set wfds;

		struct timeval tv = {0, 0};

		int err;
		
		int need_expose=0; 
		int min_x=screen.x, min_y=screen.y, max_x=0, max_y=0;
		
		while (ggiEventPoll(vis, WANTED_EVENTS, &tv) && !quit) {

			ggi_event ev;

			do_blank(0);

			ggiEventRead(vis, &ev, WANTED_EVENTS);

			switch (ev.any.type) {

			case evKeyPress:
			case evKeyRepeat:
				handle_keyboard(ev.key.sym,
						ev.key.modifiers, 
						master_fd);
				break;

			case evExpose:
				need_expose = 1;

				min_x = MIN(min_x, ev.expose.x / font.x);
				min_y = MIN(min_y, ev.expose.y / font.y);

				max_x = MAX(max_x, (ev.expose.x+ev.expose.w +
					    font.x - 1) / font.x);
				max_y = MAX(max_y, (ev.expose.y+ev.expose.h +
					    font.y - 1) / font.y);
			}

			tv.tv_sec = tv.tv_usec = 0;
		}

		if (quit) break;

		if (need_expose) {
			do_expose(min_x, min_y, max_x-min_x, max_y-min_y);
		}

		FD_ZERO(&rfds);
		FD_ZERO(&wfds);
		FD_SET(master_fd, &rfds);

		if (write_pending) {
			FD_SET(master_fd, &wfds);
		}

		tv.tv_sec = 0;
		tv.tv_usec = SLEEP_MS * 1000;

		do {
			err = select(master_fd+1, &rfds, &wfds, NULL, &tv);

		} while ((err < 0) && (errno = EINTR) && !quit);

		if (quit) break;

		if (err < 0) {
			perror("nixterm: select");
			continue;
		}

		if (FD_ISSET(master_fd, &rfds)) {
			do_blank(0);
			handle_incoming_data(master_fd);
		}

		if (FD_ISSET(master_fd, &wfds)) {
			do_blank(0);
			handle_outgoing_data(master_fd);
		}

		if ((blank_time > 0) && ! blanked) {

			struct timeval cur;
			int millies;

			gettimeofday(&cur, NULL);

			millies = (cur.tv_sec-blank_start.tv_sec)*1000 + 
				(cur.tv_usec-blank_start.tv_usec)/1000;

			if (millies > blank_time) {
				do_blank(1);
			}
		}
	}

	blacken_whole_screen();
	
	ggiClose(vis);
	ggiExit();

	switch (quit) {
		case 1: fprintf(stderr, "Exiting... (user).\n\n");
			break;
		case 2: fprintf(stderr, "Exiting... (shell).\n\n");
			break;
		case 3: fprintf(stderr, "Exiting... (pipe error).\n\n");
			break;
	}
	
	exit(0);
}
